/* exception.h --- header implementing lightweight, but featured
                   exception handling.

Time-stamp: "2003-08-09 21:34:18 admp"

Copyright (C) 2003 admp <admp@licejus.lt>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA. */

/* Implementation in 444 lines.

Implemented:
  1. TRY --- a catcher block.
  2. EXCEPT, EXPECT, CATCH --- a handler block.
  3. RAISE --- raises an exception.
  4. ON --- a handler.

Syntax:
  1. TRY
  The syntax is:
    try { statements; } except { handler blocks; }

  2. EXCEPT, EXPECT, CATCH
  The syntax is:
    except { handler blocks; }

  3. RAISE
  The syntax is: 
    raise (params)
  PARAMS depend on your local implementation.  By default, it is a
  number (int).
  
  4. ON
  The syntax is:
    on (params) { statements; }
  PARAMS are the same as for RAISE.

Supports:
  1. Nested catcher blocks.
  2. Raising an exception outside catcher block (this results in an 
     unhandled exception.)
  3. Raising an exception in function called from catcher block.
  4. Different types for exception structures.
  5. Reporting file and line where exception was raised/handled. */

/* Search for TODO and FIXME tags if you want to enchance this.
   (This would be nice.) */

#if defined __EXCEPTION_H__
  /* TODO: allow compile as whole-in-one, header OR source file. */
#  error Header file "exception.h" used twice.  This is the Wrong Thing.
#endif

#define __EXCEPTION_H__

#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* The default prefix is `__exc' for functions and `__EXC' for
   macros. */



/* C syntax hacks.  */

/* Execute START, then block after the macro, and finally END. */

#define __EXC_BLOCK(start, end)              \
  for (start, __exc_block_pass = 1;          \
       __exc_block_pass;                     \
       end, __exc_block_pass = 0)


/* Likewise, but START block is empty. */

#define __EXC_END(end)                       \
  for (__exc_block_pass = 1;                 \
       __exc_block_pass;                     \
       end, __exc_block_pass = 0)



/* The exception value. */


/* You'll want to make local changes to these.  For example, to use
   your own exception structure. */

#ifndef __EXC_TYPE
#  define __EXC_TYPE               int

/* Include the default __EXC_PRINT. */
#  define __EXC_TYPE_DEFAULT
#endif

#ifndef __EXC_MAKE
#  define __EXC_MAKE(code...)      code
#endif

#ifndef __EXC_EQ
#  define __EXC_EQ(c1, c2)         ((c1) == (c2))
#endif


/* Optional exception printer.  This is used for debugging purposes
   only.  */

#if !defined (__EXC_PRINT) && defined (__EXC_TYPE_DEFAULT)
#  define __EXC_PRINT(e, stream)             \
     fprintf (stream, "%d", e)
#endif



/* All variables are declared volatile to force non-optimization. */

/* This counter is used by __EXC_BLOCK.  It works well even if nested. */

volatile int __exc_block_pass;

/* Flag to be set by ON? */

volatile int __exc_handled;

/* For indexing every call to TRY. */

volatile unsigned __exc_tries;

/* These identify the raised exception. */

char *__exc_file;
unsigned __exc_line;
__EXC_TYPE __exc_code;

struct __exc_stack
{
  unsigned num;
  jmp_buf j;
  struct __exc_stack *prev;
};

/* This is the global stack of catchers. */
struct __exc_stack *__exc_global;



/* Debugging of exceptions.  Nothing interesting for you.  (Just for me.) */

#ifdef __EXC_DEBUG

#  include <stdarg.h>

#  ifndef __EXC_STREAM
/* I often redirect debugging information to a file. */
#    define __EXC_STREAM                     stdout
#  endif

/* Prints error message. */

static void
__exc_debug (char *fmt, ...)
{
  va_list ap;

  fprintf (__EXC_STREAM, "__EXC: ");
  va_start (ap, fmt);
  vfprintf (__EXC_STREAM, fmt, ap);
  va_end (ap);
  fprintf (__EXC_STREAM, "\n");
}

/* For printing __exc_global. */

static void
__exc_print_global ()
{
  struct __exc_stack *level = __exc_global;
  unsigned items;

  if (level == NULL)
    {
      fprintf (__EXC_STREAM, "Stack empty\n");
      return;
    }

  fprintf (__EXC_STREAM, "Current stack (from bottom to top):\n");
  for (items = 0; level; level = level->prev)
    {
      fprintf (__EXC_STREAM, "%c ", items == 0 ? '[' : ' ');

      /* FIXME: does printing jmp_buf literally (as we now address and
	 its size) good?  No, if human cannot understand it.  One
	 can't unless he chooses one popular C library and study the
	 implementation carefully;) */
      fprintf (__EXC_STREAM, "%u", level->num);
      /* TODO: newline every four or so items.  Use tab stops. */
      fprintf (__EXC_STREAM, " %c\n", level->prev ? ' ' : ']');  
      items++;
    }

  fprintf (__EXC_STREAM, "Totally %u items.\n", items);
}

#else
#  define __exc_debug(args...)
#  define __exc_print_global()
#endif



/* Pop exception from stack, putting into J (if nonzero).  If stack is
   empty, print error message and exit.  Used in EXCEPT. */

static void
__exc_pop (jmp_buf *j)
{
  register struct __exc_stack *stored = __exc_global;

  __exc_debug ("POP () to %p", j);

  if (stored == NULL)
    {
      __exc_debug ("Unhandled exception.");

      fprintf (stderr, "Unhandled exception:\n");
      fprintf (stderr, "At file \"%s\", line %u.\n", 
	       __exc_file, __exc_line);

#ifdef __EXC_PRINT
      /* Try to print exception itself. */
      fprintf (stderr, "Exception:\n");
      __EXC_PRINT (__exc_code, stderr);
      fprintf (stderr, "\n");
#endif

      exit (3);
    }

  __exc_global = stored->prev;

  if (j)
    {
      /* This assumes that JMP_BUF is a structure etc. and can be
	 copied rawely.  This is true in GLIBC, as I know. */
      memcpy (j, &stored->j, sizeof (jmp_buf));
    }

  __exc_debug ("Popped");
  __exc_print_global ();

  /* While with MALLOC, free.  When using obstacks it is better not to
     free and hold up. */
  free (stored);
}



/* Push J onto the stack, with RETURNED as value from SETJMP.  Return
   nonzero, if RETURNED is 0.  If RETURNED is nonzero, returns 0.
   Used in TRY. */

static int
__exc_push (jmp_buf *j, int returned)
{
  struct __exc_stack *new;

  __exc_debug ("PUSH (), %p, %d", j, returned);

  /* SETJMP returns 0 first time, nonzero from __EXC_RAISE. */
  if (returned != 0)
    {
      __exc_debug ("Returning from RAISE?");
      return 0;
    }

  /* Since this didn't come from RAISE, fine to increase counter. */
  ++__exc_tries;
  __exc_debug ("This is PUSH () number %u", __exc_tries);

  new = malloc (sizeof (struct __exc_stack));
  memcpy (&new->j, j, sizeof (jmp_buf));
  new->num = __exc_tries;
  new->prev = __exc_global;
  __exc_global = new;

  __exc_print_global ();

  return 1;
}



/* Raise an exception in FILE at LINE, with code CODE.  Used in RAISE. */

static void
__exc_raise (char *file, unsigned line, __EXC_TYPE code)
{
  jmp_buf j;

  __exc_debug ("RAISE ()");
  __exc_debug ("An exception raised in file \"%s\" at line %u",
	       file, line);

#if (defined __EXC_DEBUG) && (defined __EXC_PRINT)
  __EXC_PRINT (code, __EXC_STREAM);
  fprintf (__EXC_STREAM, "\n");
#endif

  __exc_file = file;
  __exc_line = line;
  __exc_code = code;

  /* Pop for jumping. */
  __exc_pop (&j);
  
  __exc_debug ("Jumping to the handler");

  /* LONGJUMP to J with nonzero value. */
  longjmp (j, 1);
}

/* Raise it in upper level of catcher blocks. */

static void
__exc_reraise ()
{
  jmp_buf j;

  __exc_debug ("RERAISE ()");
  __exc_debug ("An exception was raised in file \"%s\" at line %u",
	       __exc_file, __exc_line);

#if (defined __EXC_DEBUG) && (defined __EXC_PRINT)
  __EXC_PRINT (__exc_code, stderr);
  fprintf (__EXC_STREAM, "\n");
#endif

  __exc_pop (&j);
  longjmp (j, 1);
}



/* What a f...  Somewhy I can't get GCC's __attribute__ working here
   to tell that FILE and LINE are unused in non-debuging mode. */

/* TODO: define __attribute__ (foo) to do nothing if this is not GCC.
   (Check if GNUC is predefined.) */

static int
__exc_on (char *file, unsigned line, __EXC_TYPE code)
{
  /* It is impossible to jump to the end of EXCEPT block, nor it is
     possible to mix ELSE with IF, so use before-defined __EXC_HANDLED
     flag. 

     Actually, use of
      on (...) ...;
      else on (...) ...;

     Is possible, but don't force the *user* to care about it. */

  __exc_debug ("ON ()");
  __exc_debug ("Trying to handle in file \"%s\", line %u", file, line);

  if (__exc_handled == 1)
    {
      __exc_debug ("Exception already handled in this level, skip");
      return 0;
    }
  
  if (__EXC_EQ (code, __exc_code))
    {
      __exc_debug ("This handler FITS");

      __exc_handled = 1;
      return 1;
    }

  __exc_debug ("This handler DOESN'T FIT");

  /* Not matched. */
  return 0;
}



/* Start catching. */

/* Obviously, there is no way to check if appropriate EXCEPT exists.
   Its non-exsistence won't do segfault etc.; program will simply do
   the thing after TRY, without any error handling.  Raising from
   there works. */

#define try                                  \
  if (({jmp_buf __exc_j;                     \
        int __exc_ret;                       \
        __exc_ret = setjmp (__exc_j);        \
        __exc_push (&__exc_j, __exc_ret);})) \
    __EXC_END(__exc_pop (0))

#define raise(code...)                       \
  __exc_raise (__FILE__, __LINE__,           \
               __EXC_MAKE (code))

/* RAISE in EXCEPT block won't go into itself, because corresping item
   from __EXC_GLOBAL was already popped. */

#define except                               \
  else                                       \
    __EXC_BLOCK (__exc_handled = 0,          \
                 ({ if (__exc_handled == 0)  \
                     __exc_reraise (); }))

/* EXPECT is an alias for EXCEPT. */

#define expect                         except

/* CATCH is an alias for EXCEPT. */

#define catch                          except

/* Try to handle an exception. */

#define on(code...)                          \
  if (__exc_on (__FILE__, __LINE__,          \
                __EXC_MAKE (code)))

/* End. */
